#include <stdio.h>
#include <vector>
#include <unistd.h>
#include <string.h>
#include <mpi.h>
#include <algorithm>
#include <string>
#include <iostream>


using namespace std;

class MPI_Master_Slave
{
private:
    int _master;
    int _rank;
    int _size;
    int _iArrSize;
    int _iPos;
    int _iTotalPos;
    bool _bEnd;
    int flag;
    
    vector<char> _iMessages;
    vector<int> _SleepTimes;
    vector<bool> _Listening;
    vector<MPI_Status> _stat;
    vector<MPI_Request> _req;
    vector<char> _iLetters;
public:
    MPI_Master_Slave(int rank, int size, vector<MPI_Status> &stat, vector<MPI_Request> &req);
    ~MPI_Master_Slave() = default;
    void Prepare_Data();
    void Notify_Master();
    void Notify_Slaves();
    void Master_Listen();
    void Slave_Listen();
    void Listen();
    void Go();
    void Give_Work(int rank);
};

MPI_Master_Slave::MPI_Master_Slave(int rank, int size, vector<MPI_Status> &stat, vector<MPI_Request> &req)
{
    _stat = stat;
    _req = req;
    _rank = rank;
    _master = 0;
    _size = size;
    flag = 0;
    _SleepTimes.resize(_size);
    _iMessages.resize(_size);
    _iPos = 0;
    _iTotalPos = 0;
    _Listening = vector<bool>(_size,false);
    _bEnd = false;
    if(_rank == _master)
    {
        for(int k = 0; k<_size;k++)
        {
            if(k==0)
                _SleepTimes[k] = 1;
            else  
                _SleepTimes[k] = 2;
        }
        Prepare_Data();
        _iArrSize = _iLetters.size();
    }
    MPI_Bcast(_SleepTimes.data(),_size,MPI_INT,_master,MPI_COMM_WORLD);
    MPI_Bcast(&_iArrSize,1,MPI_INT,_master,MPI_COMM_WORLD);
    if(_rank != _master)
    {
        _iLetters.resize(_iArrSize);
    }
    MPI_Bcast(_iLetters.data(),_iArrSize,MPI_CHAR,_master,MPI_COMM_WORLD);
    if(_rank == _master)
        printf("Data Broadcasted\n");
    Go();
}

void MPI_Master_Slave::Prepare_Data()
{

    vector<char> tmpLetters = {'A','A','A','B','B','C','A','A','A','B','B','B','C','D','A','A','A','B','A','A','A','B','B','B','C','A','A','A','B','B','B','C','D','A','A','A','B','A','E','A','A','B','B','B','C','A','A','A','B','B','B','C','D','A','A','A','B','E','B','C'};
    _iLetters = tmpLetters;
    string sOutput;
    sOutput += "New data are: \n";
    for(int k = 0; k<_size;k++)
    {
        sOutput+= to_string(k) + " - " + to_string(_SleepTimes[k]) + "\n"; 
    }
    cout<<sOutput<<endl;
    string sOutput2;
    cout<< "Letters are: ";
    for(int k = 0; k<_iLetters.size();k++)
    {
        cout<< _iLetters[k]<< " "; 
    }
    cout<<endl;

}

void MPI_Master_Slave::Master_Listen()
{
    for(int slave=0;slave<_size;slave++)
    {
        if(!_Listening[slave])
        {
            MPI_Irecv(&_iMessages[slave],1,MPI_CHAR,slave,MPI_ANY_TAG,MPI_COMM_WORLD,&_req[slave]);
            _Listening[slave] = true;
        }
    }
}

void MPI_Master_Slave::Slave_Listen()
{
    MPI_Irecv(&_iMessages[_rank],1,MPI_CHAR,_master,MPI_ANY_TAG,MPI_COMM_WORLD,&_req[_rank]);

}

void MPI_Master_Slave::Listen()
{
    if(_rank == _master)
    {
        for(int slave = 0; slave<_size;slave++)
        {
            MPI_Test(&_req[slave],&flag,&_stat[slave]);
            if(flag==0)
            {
                continue;
            }
            else
            {
                _Listening[slave] = false;
                char cMessage = _iMessages[slave];
                if(cMessage != 'E')
                {
                    int iNewPos = _iTotalPos++;
                    MPI_Send(&cMessage,1,MPI_CHAR,slave,0,MPI_COMM_WORLD);
                    MPI_Send(&iNewPos,1,MPI_INT,slave,0,MPI_COMM_WORLD);
                }
                else
                {
                    int iNewPos = _iArrSize;
                    _bEnd = true;
                    MPI_Send(&cMessage,1,MPI_CHAR,slave,0,MPI_COMM_WORLD);
                    MPI_Send(&iNewPos,1,MPI_INT,slave,0,MPI_COMM_WORLD);
                }

            }
        }
        if(_bEnd==true)
        {
            for(int slave = 0; slave<_size;slave++)
            {
                char cLetter = 'E';
                if(slave == _master)
                    continue;
                int iNewPos = _iArrSize;
                MPI_Isend(&cLetter,1,MPI_CHAR,slave,0,MPI_COMM_WORLD,&_req[slave]);
                MPI_Send(&iNewPos,1,MPI_INT,slave,0,MPI_COMM_WORLD);
            }
        }
        Master_Listen();
    }
    else
    {
        MPI_Test(&_req[_rank],&flag,&_stat[_rank]);
        if(flag==0)
        {
            return;
        }
        else
        {
            char cMessage = _iMessages[_rank];
            if(cMessage=='E')
            {
                _bEnd = true;
                MPI_Recv(&_iPos,1,MPI_INT,_master,0,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
            }
            else
            {
                MPI_Recv(&_iPos,1,MPI_INT,_master,0,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
            }
            Slave_Listen();
        }
    }
}


void MPI_Master_Slave::Go()
{
    char cActLetter;
    if(_rank == _master)
    {
        Master_Listen();
        _iPos = _iTotalPos++;
    }
    else
    {
        Slave_Listen();
        cActLetter = 'N';
        MPI_Isend(&cActLetter,1,MPI_CHAR,_master,0,MPI_COMM_WORLD,&_req[_rank]);
        MPI_Wait(&_req[_rank],&_stat[_rank]);
        do
        {
            Listen();
        } while (flag==0);

    }
    while(_iPos<_iArrSize)
    {
        sleep(_rank+1);
        Listen();
        if(_bEnd==true)
        {
            break;
        }
        cActLetter = _iLetters[_iPos];
        printf("Process %d has pos %d and char %c\n",_rank,_iPos,cActLetter);
        Listen();
        if(_rank==_master)
        {
            if(cActLetter == 'E')
                _bEnd = true;
            _iPos = _iTotalPos++;
        }
        else
        {
            MPI_Isend(&cActLetter,1,MPI_CHAR,_master,0,MPI_COMM_WORLD,&_req[_rank]);
            Listen();
        }
    }
    printf("Process %d END\n",_rank);
    if(_rank==_master)
    {
        while(!_bEnd)
            Listen();
    }
}






int main(int argc, char **argv)
{
    int rank;
    int size;
    char name[80];
    int lenght;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    char cMessage;
    vector<MPI_Request> req(size);
    vector<MPI_Status> stat(size);
    MPI_Master_Slave worker = MPI_Master_Slave( rank,  size,  stat,  req);
    MPI_Finalize();

}