#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <stddef.h>
#include <vector>

typedef struct car_s {
        char ECS;
        int price;
        short year_of_fabrication;
        float fuel_consuption;
} car;

using namespace std;

int main(int argc, char **argv) {

    const int tag = 13;
    int size, rank;

    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    if (size < 3) {
        fprintf(stderr,"Requires at least 3 processes.\n");
        exit(-1);
    }

    /* create a type for struct car */
    const int nitems=2;
    int          blocklengths[4] = {1,1,1,};
    MPI_Datatype types[4] = {MPI_CHAR, MPI_INT, MPI_SHORT,MPI_FLOAT};
    MPI_Datatype mpi_car_type;
    MPI_Aint     offsets[4];

    offsets[0] = offsetof(car, ECS);
    offsets[1] = offsetof(car, price);
    offsets[2] = offsetof(car, year_of_fabrication);
    offsets[3] = offsetof(car, fuel_consuption);

    MPI_Type_create_struct(nitems, blocklengths, offsets, types, &mpi_car_type);
    MPI_Type_commit(&mpi_car_type);

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    vector<car> cars;
    int num_cars;
    //Add path to argument when starting a programm
    if (rank == 0) {
        FILE *f = fopen(argv[1],"r");
        fscanf(f,"%d\n",&num_cars);
        for(int k = 0; k<num_cars;k++)
        {
            car tmp_car;
            char tmp_ECS;
            int tmp_price;
            int tmp_year;
            float tmp_fuel;
            fscanf(f,"%c %d %d %f\n",&tmp_ECS,&tmp_price,&tmp_year,&tmp_fuel);
            tmp_car.ECS = tmp_ECS;
            tmp_car.price = tmp_price;
            tmp_car.year_of_fabrication = (short)tmp_year;
            tmp_car.fuel_consuption = tmp_fuel;
            cars.push_back(tmp_car);
        }
        fclose(f);
    }
    MPI_Bcast(&num_cars,1,MPI_INT,0,MPI_COMM_WORLD);
    if(rank!=0)
        cars.resize(num_cars);
    MPI_Bcast(cars.data(),num_cars,mpi_car_type,0,MPI_COMM_WORLD);
    if(rank==0)
    {
        int count_c = 0;
        for(int k = 0; k<cars.size();k++)
        {
            count_c += (int)(cars[k].ECS == 'C');
        }
        printf("Cars in C class: %d\n",count_c);
    }
    if(rank==1)
    {
        vector<float> fuels(6,0.0);
        vector<int> class_counts(6,0);
        for(int k = 0 ; k<cars.size();k++)
        {
            int idx = (int)(cars[k].ECS - 'A');
            fuels[idx]+= cars[k].fuel_consuption;
            class_counts[idx]++;
        }
        char class_names[6] = {'A','B','C','D','E','F'}; 
        for(int k = 0; k<6;k++)
        {
            fuels[k]/= class_counts[k];
            printf("Class %c hase avg. fuel consumption %f l/km\n",class_names[k],fuels[k]);
        }
    }
    if(rank==2)
    {
        int price = 0;
        int count = 0;
        for(int k = 0 ; k<cars.size();k++)
        {
            if(abs((int)(cars[k].ECS - 'A'))<=2)
            {
                count++;
                price += cars[k].price;
            }
        }
        float avg_price = ((float)price)/((float)count);
        printf("Average price for ECS: A B and C cars is %0.3f CZK\n",avg_price);
    }
    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Type_free(&mpi_car_type);
    MPI_Finalize();

    return 0;
}